from __future__ import annotations

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from tenacity import (
    retry,
    stop_after_attempt,
    wait_exponential,
    before_log,
    after_log,
    before_sleep_log,
)

from ..file import S3Object
from ..operation import Operation
from ..utils.decorators import supported_filetypes, with_pango_docs
from ..utils import get_random_string, monitor_retry_condition
from ..utils.exceptions import SkippedOperation
from ..utils.s3 import calculate_etag, TransferConfig, S3ProgressPercentage

import logging
from pathlib import Path

logger = logging.getLogger(__name__)


@supported_filetypes(filetypes=[S3Object])
@with_pango_docs(filename="s3_downloader.pango")
class S3DownloaderOperation(Operation):

    NAME = "S3 Downloader"

    def __init__(self, *args, **kwargs):
        Operation.__init__(self, *args, **kwargs)
        self._grid = Gtk.Grid(
            border_width=5,
            row_spacing=5,
            column_spacing=5,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        self.add(self._grid)

        # We are going to use the client from the engine
        # We only need to specify a directory where to save the files locally
        label = Gtk.Label(
            label="Download Destination",
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        self._grid.attach(label, 0, 0, 1, 1)

        self._directory_chooser_button = self.register_widget(
            Gtk.FileChooserButton(
                title="Select the directory to write to",
                action=Gtk.FileChooserAction.SELECT_FOLDER,
                create_folders=True,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.FILL,
                hexpand=True,
                vexpand=False,
            ),
            "download_destination",
        )
        self._grid.attach(self._directory_chooser_button, 1, 0, 1, 1)

    def preflight_check(self):

        # to check:
        # 1. Does the engine export the S3 client?
        # 2. Can we list the bucket?
        # 3. Is the destination writable?

        # write random file into destination folder
        temp_filename = Path(
            self.params.download_destination, get_random_string(6) + ".dat"
        )
        temp_filename.write_text("delete me")
        temp_filename.unlink()

    @retry(
        retry=monitor_retry_condition(),
        reraise=True,
        stop=stop_after_attempt(5),
        wait=wait_exponential(),
        before=before_log(logger, logging.DEBUG),
        after=after_log(logger, logging.DEBUG),
        before_sleep=before_sleep_log(logger, logging.DEBUG),
    )
    def run(self, file: S3Object):  # type: ignore[override]

        s3_client = self.appwindow.active_engine.s3_client
        bucket_name = self.appwindow.active_engine._get_params().bucket_name

        # Check if file already exists in the destination folder
        destination = Path(
            self.params.download_destination, *file.relative_filename.parts
        )

        # get size and etag
        response = s3_client.head_object(
            Bucket=bucket_name,
            Key=file.key,
        )
        key_size = int(response["ContentLength"])

        if destination.exists():
            if destination.stat().st_size == key_size:
                remote_etag = response["ETag"][
                    1:-1
                ]  # get rid of those extra quotes
                # note: for this to work with objects that were initially
                # uploaded with multipart uploads,
                # then the upload must have used the same TransferConfig as used in S3UploaderOperation
                local_etag = calculate_etag(destination)
                if remote_etag == local_etag:
                    # attach metadata
                    raise SkippedOperation("File has been downloaded already")
        else:
            # ensure the parent directories exist
            destination.parent.mkdir(parents=True, exist_ok=True)

        try:
            s3_client.download_file(
                Bucket=bucket_name,
                Key=file.key,
                Filename=str(destination),
                Config=TransferConfig,
                Callback=S3ProgressPercentage(
                    file, destination, self.index, float(key_size)
                ),
            )
        except Exception as e:
            logger.exception(f"S3UploaderOperation.run exception")
            return f"Could not download {str(destination)}: {str(e)}"

        return None



from rfi_file_monitor.utils import ExitableThread
import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

import time
from threading import current_thread
import logging
from random import random

from ..operation import Operation
from ..utils.exceptions import SkippedOperation
from ..utils.decorators import (
    with_pango_docs,
    supported_filetypes,
    add_directory_support,
)
from ..file import File, RegularFile, S3Object, URL, Directory

logger = logging.getLogger(__name__)


@with_pango_docs(filename="dummy_operation.pango")
@supported_filetypes(filetypes=[RegularFile, S3Object, URL, Directory])
class DummyOperation(Operation):

    NAME = "Dummy Operation"
    DEBUG = True

    def __init__(self, *args, **kwargs):
        Operation.__init__(self, *args, **kwargs)
        self._grid = Gtk.Grid(
            row_spacing=5,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        self.add(self._grid)
        self._grid.attach(
            Gtk.Label(label="This is a dummy operation"), 0, 0, 1, 1
        )
        combobox = Gtk.ComboBoxText()
        combobox.append_text("Text1")
        combobox.append_text("Text2")
        combobox.append_text("Text3")
        combobox.set_active(0)
        self.register_widget(combobox, "dummy_combo")
        self._grid.attach(combobox, 0, 1, 2, 1)

        widget = self.register_widget(
            Gtk.CheckButton(
                active=False,
                label="Use bucket tags",
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "enable_bucket_tags",
        )
        self._grid.attach(widget, 0, 2, 2, 1)

        widget = self.register_widget(
            Gtk.CheckButton(
                active=False,
                label="Use object tags",
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "enable_object_tags",
        )
        self._grid.attach(widget, 0, 3, 2, 1)

        widget = self.register_widget(
            Gtk.CheckButton(
                active=False,
                label="Test Echo ACL",
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "enable_echo_acl",
        )
        self._grid.attach(widget, 0, 4, 2, 1)

        widget = self.register_widget(
            Gtk.CheckButton(
                active=False,
                label="Fail randomly",
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "enable_random_fails",
        )
        self._grid.attach(widget, 0, 5, 2, 1)

        widget = self.register_widget(
            Gtk.CheckButton(
                active=False,
                label="Skip randomly",
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "enable_random_skips",
        )
        self._grid.attach(widget, 0, 6, 2, 1)

        widget = self.register_widget(
            Gtk.CheckButton(
                active=False,
                label="Process as directories if supported",
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "enable_directories",
        )
        self._grid.attach(widget, 0, 7, 2, 1)

        widget = self.register_widget(
            Gtk.CheckButton(
                active=False,
                label="Add Operational Metadata",
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "add_metadata",
        )
        self._grid.attach(widget, 0, 8, 2, 1)

    def preflight_check(self):
        metadata = dict()

        if self.params.enable_echo_acl:
            metadata["bucket_acl_options"] = {
                "AccessControlPolicy": {
                    "Grants": [
                        {
                            "Grantee": {
                                "ID": "rfi-ai",
                                "Type": "CanonicalUser",
                            },
                            "Permission": "FULL_CONTROL",
                        },
                        {
                            "Grantee": {
                                "ID": "rfi-instrument-xevo",
                                "Type": "CanonicalUser",
                            },
                            "Permission": "FULL_CONTROL",
                        },
                    ],
                    "Owner": {
                        "ID": "rfi-instrument-xevo",
                    },
                },
            }
            metadata["object_acl_options"] = metadata["bucket_acl_options"]

        if self.params.enable_bucket_tags:
            metadata["bucket_tags"] = {
                "owner": "Tom",
                "group": "Toms friends",
                "experiment name": "Black magic part1",
            }

        if self.params.enable_object_tags:
            metadata["object_tags"] = {
                "owner": "Laura",
                "group": "Lauras friends",
                "experiment name": "Alchemy part2",
            }

        if metadata:
            self.appwindow.preflight_check_metadata[self.index] = metadata

    def run(self, file: File):
        if isinstance(file, Directory) and self.params.enable_directories:
            return add_directory_support(type(self)._run)(self, file)
        return self._run(file)

    @classmethod
    def _attach_metadata(cls, file: File, operation_index: int):
        file.operation_metadata[operation_index] = {"number": random()}

    def _run(self, file: File):
        logger.info(f"Processing {file.filename}")
        thread = current_thread()
        for i in range(10):
            if isinstance(thread, ExitableThread) and thread.should_exit:
                return str("Thread killed")
            time.sleep(1.0)

            if self.params.enable_random_fails and random() < 0.02:
                return "Unfavorable RNG!!!"
            elif self.params.enable_random_skips and random() < 0.02:
                raise SkippedOperation("Unfavorable RNG!")

            file.update_progressbar(self.index, (i + 1) * 10)
            if self.params.add_metadata:
                self._attach_metadata(file, self.index)
        # None indicates success, a string failure, with its contents set to an error message
        return None



from __future__ import annotations

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib
from ..utils import match_path, _get_common_patterns
from watchdog.observers import Observer
from watchdog.events import (
    FileSystemEventHandler,
    DirCreatedEvent,
    FileCreatedEvent,
    FileModifiedEvent,
)

from .directory_watchdog_engine_advanced_settings import (
    DirectoryWatchdogEngineAdvancedSettings,
)
from ..engine import Engine, EngineThread
from ..file import Directory, FileStatus
from ..utils import (
    get_file_creation_timestamp,
    LongTaskWindow,
    get_patterns_from_string,
)
from ..utils.decorators import (
    exported_filetype,
    with_advanced_settings,
    with_pango_docs,
)

import logging
from typing import List
from pathlib import Path, PurePath
import os

logger = logging.getLogger(__name__)

ERROR_MSG = "Ensure that the selected directory is readable and that any provided patterns do not conflict"


@with_pango_docs(filename="directory_watchdog_engine.pango")
@with_advanced_settings(
    engine_advanced_settings=DirectoryWatchdogEngineAdvancedSettings
)
@exported_filetype(filetype=Directory)
class DirectoryWatchdogEngine(Engine):

    NAME = "Directories Monitor"

    def __init__(self, appwindow):
        super().__init__(appwindow, DirectoryWatchdogEngineThread, ERROR_MSG)

        label = Gtk.Label(
            label="Monitored Directory",
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        self.attach(label, 0, 0, 1, 1)

        self._directory_chooser_button = self.register_widget(
            Gtk.FileChooserButton(
                title="Select a directory for monitoring",
                action=Gtk.FileChooserAction.SELECT_FOLDER,
                create_folders=True,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.FILL,
                hexpand=True,
                vexpand=False,
            ),
            "monitored_directory",
        )
        self.attach(self._directory_chooser_button, 1, 0, 1, 1)
        self._directory_chooser_button.connect(
            "selection-changed", self._directory_chooser_button_cb
        )

        self._monitor: Observer = None

    def _directory_chooser_button_cb(self, button):
        if (
            self.params.monitored_directory is None
            or Path(self.params.monitored_directory).is_dir() is False
        ):
            self._valid = False
        else:
            try:
                os.listdir(self.params.monitored_directory)
                self._valid = True
            except Exception:
                self._valid = False
        self.notify("valid")


# add our ExitableThread methods to Watchdog's Observer class to ensure it can be used as an EngineThread
class DirectoryWatchdogEngineThread(Observer):
    def __init__(
        self, engine: DirectoryWatchdogEngine, task_window: LongTaskWindow
    ):
        super().__init__()
        self._should_exit: bool = False
        self._engine: Engine = engine
        self.params = engine.params
        self._task_window = task_window
        app = engine.appwindow.props.application

        self._included_file_patterns = app.get_allowed_file_patterns(
            self.params.allowed_file_patterns
        )
        self._excluded_file_patterns = app.get_ignored_file_patterns(
            self.params.ignore_file_patterns
        )
        self._included_directory_patterns = get_patterns_from_string(
            self.params.allowed_directory_patterns
        )
        self._excluded_directory_patterns = get_patterns_from_string(
            self.params.ignore_directory_patterns, defaults=[]
        )

        self.schedule(
            event_handler=EventHandler(engine),
            path=engine.params.monitored_directory,
            recursive=True,
        )

    @property
    def should_exit(self):
        return self._should_exit

    @should_exit.setter
    def should_exit(self, value: bool):
        self._should_exit = value
        if self._should_exit:
            self.stop()

    def _search_for_existing_files(self, directory: Path) -> int:
        rv: int = 0
        for child in directory.iterdir():
            if (
                child.is_file()
                and not child.is_symlink()
                and match_path(
                    child,
                    included_patterns=self._included_file_patterns,
                    excluded_patterns=self._excluded_file_patterns,
                    case_sensitive=False,
                )
            ):

                rv += 1
            elif child.is_dir() and not child.is_symlink():
                rv += self._search_for_existing_files(directory.joinpath(child))
        return rv

    def _dir_filter(self, child: Path) -> bool:
        return (
            child.is_dir()
            and not child.is_symlink()
            and match_path(
                child,
                included_patterns=self._included_directory_patterns,
                excluded_patterns=self._excluded_directory_patterns,
                case_sensitive=False,
            )
            and (self._search_for_existing_files(child) > 0)
        )

    def _search_for_existing_directories(
        self, directory: Path
    ) -> List[Directory]:
        rv: List[Directory] = list()
        dirs = filter(self._dir_filter, directory.iterdir())

        for _dir in dirs:
            rv.append(
                Directory(
                    str(_dir),
                    PurePath(_dir.name),
                    get_file_creation_timestamp(_dir),
                    FileStatus.SAVED,
                    self._included_file_patterns,
                    self._excluded_file_patterns,
                )
            )
        return rv

    def run(self):
        # confirm patterns are valid
        if bool(
            common_file_patterns := _get_common_patterns(
                self._included_file_patterns,
                self._excluded_file_patterns,
                False,
            )
        ):
            self._engine.cleanup()
            GLib.idle_add(
                self._engine.abort,
                self._task_window,
                f"Common file patterns {common_file_patterns} detected!",
                priority=GLib.PRIORITY_HIGH,
            )
            return

        if bool(
            common_directory_patterns := _get_common_patterns(
                self._included_directory_patterns,
                self._excluded_directory_patterns,
                False,
            )
        ):
            self._engine.cleanup()
            GLib.idle_add(
                self._engine.abort,
                self._task_window,
                f"Common directory patterns {common_directory_patterns} detected!",
                priority=GLib.PRIORITY_HIGH,
            )
            return

        if self.params.process_existing_directories:
            GLib.idle_add(
                self._task_window.set_text,
                "<b>Processing existing directories...</b>",
            )
            try:
                existing_directories = self._search_for_existing_directories(
                    Path(self.params.monitored_directory)
                )
                GLib.idle_add(
                    self._engine._appwindow._queue_manager.add,
                    existing_directories,
                    priority=GLib.PRIORITY_DEFAULT_IDLE,
                )
            except Exception as e:
                self._engine.cleanup()
                GLib.idle_add(
                    self._engine.abort,
                    self._task_window,
                    e,
                    priority=GLib.PRIORITY_HIGH,
                )
                return

        # if we get here, things should be working.
        # close task_window
        GLib.idle_add(
            self._engine.kill_task_window,
            self._task_window,
            priority=GLib.PRIORITY_HIGH,
        )

        # kick off watchdog's run method
        super().run()

    def on_thread_stop(self):
        super().on_thread_stop()
        self._engine.cleanup()


EngineThread.register(DirectoryWatchdogEngineThread)


class EventHandler(FileSystemEventHandler):
    def __init__(self, engine: DirectoryWatchdogEngine):
        self._engine = engine
        self.params = engine.params
        app = engine.appwindow.props.application
        self._included_file_patterns = app.get_allowed_file_patterns(
            self.params.allowed_file_patterns
        )
        self._excluded_file_patterns = app.get_ignored_file_patterns(
            self.params.ignore_file_patterns
        )
        self._included_directory_patterns = get_patterns_from_string(
            self.params.allowed_directory_patterns
        )
        self._excluded_directory_patterns = get_patterns_from_string(
            self.params.ignore_directory_patterns, defaults=[]
        )
        super().__init__()

        self._empty_directories: List[Directory] = []

    def dispatch(self, event):

        paths = []
        if hasattr(event, "dest_path"):
            paths.append(os.fsdecode(event.dest_path))
        if event.src_path:
            paths.append(os.fsdecode(event.src_path))

        for path in paths:
            _path = Path(path)
            if (
                _path.is_dir()
                and match_path(
                    _path,
                    included_patterns=self._included_directory_patterns,
                    excluded_patterns=self._excluded_directory_patterns,
                    case_sensitive=False,
                )
            ) or (
                match_path(
                    _path,
                    included_patterns=self._included_file_patterns,
                    excluded_patterns=self._excluded_file_patterns,
                    case_sensitive=False,
                )
            ):

                super().dispatch(event)

    def on_created(self, event):
        path = event.src_path
        logger.info(f"Monitor found {path} for event type CREATED")
        path_object: PurePath = PurePath(path)
        rel_path = path_object.relative_to(self.params.monitored_directory)

        if isinstance(event, DirCreatedEvent):
            # if this happens directly within the monitored directory,
            # add path to the internal list
            # else if path is already in self._empty_directories that contains this, do nothing
            # else send a saved event for path, even though it may not exist in the queue!

            if (
                len(rel_path.parts) > 1
                and rel_path.parts[0] not in self._empty_directories
                and self._engine.props.running
                and self._engine._appwindow._queue_manager.props.running
            ):

                GLib.idle_add(
                    self._engine._appwindow._queue_manager.saved,
                    path,
                    priority=GLib.PRIORITY_HIGH,
                )

            elif rel_path.parts[0] not in self._empty_directories:
                logger.info(f"New directory {rel_path.parts[0]} was created")
                self._empty_directories.append(rel_path.parts[0])

        elif isinstance(event, FileCreatedEvent):
            # if the corresponding toplevel directory is in the internal array, it's time to send it to the queue
            # else, this should trigger a saved change of the corresponding Directory instance.
            # ignore if this file is put directly in the monitored directory
            if len(rel_path.parts) == 1:
                logger.info(f"Ignoring file in monitored directory {path}")
                return

            if rel_path.parts[0] in self._empty_directories:
                self._empty_directories.remove(rel_path.parts[0])
                new_path = os.path.join(
                    self.params.monitored_directory, rel_path.parts[0]
                )
                creation_timestamp = get_file_creation_timestamp(new_path)
                if creation_timestamp:
                    _dir = Directory(
                        new_path,
                        PurePath(rel_path.parts[0]),
                        creation_timestamp,
                        FileStatus.CREATED,
                        self._included_file_patterns,
                        self._excluded_file_patterns,
                    )
                else:
                    logger.info(f"File Not found, {path} has been skipped")
                    return
                GLib.idle_add(
                    self._engine._appwindow._queue_manager.add,
                    _dir,
                    priority=GLib.PRIORITY_HIGH,
                )

            elif (
                self._engine.props.running
                and self._engine._appwindow._queue_manager.props.running
            ):

                new_path = os.path.join(
                    self.params.monitored_directory, rel_path.parts[0]
                )
                GLib.idle_add(
                    self._engine._appwindow._queue_manager.saved,
                    new_path,
                    priority=GLib.PRIORITY_HIGH,
                )

        else:
            raise NotImplementedError(
                f"Unknown event type {str(event)} in on_created"
            )

    def on_modified(self, event):
        path = event.src_path
        logger.info(f"Monitor found {path} for event type MODIFIED")
        path_object: PurePath = PurePath(path)
        rel_path = path_object.relative_to(self.params.monitored_directory)

        if isinstance(event, FileModifiedEvent):
            # this should trigger a saved change of the corresponding Directory instance.
            # ignore if this file is put directly in the monitored directory
            if len(rel_path.parts) == 1:
                logger.info(f"Ignoring file in monitored directory {path}")
                return

            elif (
                self._engine.props.running
                and self._engine._appwindow._queue_manager.props.running
            ):
                new_path = os.path.join(
                    self.params.monitored_directory, rel_path.parts[0]
                )
                GLib.idle_add(
                    self._engine._appwindow._queue_manager.saved,
                    new_path,
                    priority=GLib.PRIORITY_HIGH,
                )



from __future__ import annotations

import boto3
import botocore

from ..utils.decorators import supported_filetypes, with_pango_docs
from ..utils.exceptions import SkippedOperation
from ..utils.s3 import S3ProgressPercentage, TransferConfig
from ..file import S3Object
from .s3_uploader import S3UploaderOperation, ALLOWED_OBJECT_ACL_OPTIONS

import logging

logger = logging.getLogger(__name__)


@supported_filetypes(filetypes=S3Object)
@with_pango_docs(filename="s3_copier.pango")
class S3CopierOperation(S3UploaderOperation):

    NAME = "S3 Copier"

    def run(self, file: S3Object):
        # prepare our clients
        s3_origin_client = self.appwindow.active_engine.s3_client
        bucket_origin_name = (
            self.appwindow.active_engine._get_params().bucket_name
        )

        client_destination_options = self._get_client_options(self.params)
        s3_destination_client = boto3.client("s3", **client_destination_options)

        # object creation options
        object_acl_options = self._get_dict_acl_options(
            self.appwindow.preflight_check_metadata,
            "object_acl_options",
            ALLOWED_OBJECT_ACL_OPTIONS,
        )
        # see https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.put_object_tagging
        object_tags = self._get_dict_tagset(
            self.appwindow.preflight_check_metadata, "object_tags"
        )

        # Check if file already exists in the destination bucket
        try:
            response_destination = s3_destination_client.head_object(
                Bucket=self.params.bucket_name,
                Key=file.key,
            )
        except botocore.exceptions.ClientError as e:
            # key not found, which is fine
            if int(e.response["Error"]["Code"]) != 404:
                return str(e)
        else:
            # compare etags
            remote_etag = response_destination["ETag"][1:-1]
            if remote_etag == file.etag:
                # attach metadata
                self._attach_metadata(
                    file,
                    client_destination_options["endpoint_url"],
                    file.key,
                    self.params,
                    self.index,
                )
                raise SkippedOperation(
                    "File has been copied already to the destination bucket"
                )

        try:
            s3_destination_client.copy(
                CopySource={"Bucket": bucket_origin_name, "Key": file.key},
                Bucket=self.params.bucket_name,
                Key=file.key,
                SourceClient=s3_origin_client,
                Callback=S3ProgressPercentage(
                    file, file.filename, self.index, float(file.size)
                ),
                Config=TransferConfig,
            )
            if object_tags:
                s3_destination_client.put_object_tagging(
                    Bucket=self.params.bucket_name,
                    Key=file.key,
                    Tagging=object_tags,
                )
            if object_acl_options:
                s3_destination_client.put_object_acl(
                    Bucket=self.params.bucket_name,
                    Key=file.key,
                    **object_acl_options,
                )
        except Exception as e:
            logger.exception(f"S3UploaderOperation.run exception")
            return str(e)
        else:
            # add object URL to metadata
            self._attach_metadata(
                file,
                client_destination_options["endpoint_url"],
                file.key,
                self.params,
                self.index,
            )
        return None



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from ..engine_advanced_settings import EngineAdvancedSettings
from ..engine import Engine
from ..utils import EXPAND_AND_FILL, PATTERN_PLACEHOLDER_TEXT


class CephS3BucketEngineAdvancedSettings(EngineAdvancedSettings):
    def __init__(self, engine: Engine):
        super().__init__(engine)

        self._row_counter = 0

        # Monitor existing files
        process_existing_files_grid = Gtk.Grid(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            column_spacing=5,
        )
        self.attach(process_existing_files_grid, 0, self._row_counter, 1, 1)
        self._row_counter += 1

        process_existing_files_switch = engine.register_widget(
            Gtk.Switch(
                halign=Gtk.Align.CENTER,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
                active=False,
            ),
            "process_existing_files",
        )
        process_existing_files_grid.attach(
            process_existing_files_switch, 0, 0, 1, 1
        )
        process_existing_files_grid.attach(
            Gtk.Label(
                label="Process existing files in bucket",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            1,
            0,
            1,
            1,
        )

        self._add_horizontal_separator()

        # Specify allowed file patterns
        allowed_patterns_grid = Gtk.Grid(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            column_spacing=5,
        )

        self.attach(allowed_patterns_grid, 0, self._row_counter, 1, 1)
        self._row_counter += 1
        allowed_patterns_grid.attach(
            Gtk.Label(
                label="Allowed filename patterns",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            0,
            1,
            1,
        )
        self._allowed_patterns_entry = engine.register_widget(
            Gtk.Entry(
                placeholder_text=PATTERN_PLACEHOLDER_TEXT,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "allowed_patterns",
        )
        allowed_patterns_grid.attach(self._allowed_patterns_entry, 1, 0, 1, 1)

        self._add_horizontal_separator()

        # Specify ignored file patterns
        ignore_patterns_grid = Gtk.Grid(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            column_spacing=5,
        )

        self.attach(ignore_patterns_grid, 0, self._row_counter, 1, 1)
        self._row_counter += 1
        ignore_patterns_grid.attach(
            Gtk.Label(
                label="Ignored filename patterns",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            0,
            1,
            1,
        )
        self._ignored_patterns_entry = engine.register_widget(
            Gtk.Entry(
                placeholder_text=PATTERN_PLACEHOLDER_TEXT,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "ignore_patterns",
        )
        ignore_patterns_grid.attach(self._ignored_patterns_entry, 1, 0, 1, 1)

        self._add_horizontal_separator()

        # now the RabbitMQ config
        rabbitmq_radiobutton = engine.register_widget(
            Gtk.RadioButton(
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
                active=True,
                label="AMQP 0.9.1",
            ),
            "use_rabbitmq",
        )

        rabbitmq_frame = Gtk.Frame(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            label_widget=rabbitmq_radiobutton,
        )

        self.attach(rabbitmq_frame, 0, self._row_counter, 1, 1)
        self._row_counter += 1

        rabbitmq_grid = Gtk.Grid(
            **EXPAND_AND_FILL, row_spacing=5, column_spacing=5, border_width=5
        )
        rabbitmq_frame.add(rabbitmq_grid)

        # we need:
        # 1. rabbitmq hostname
        # 2. rabbitmq username
        # 3. rabbitmq password
        # 4. rabbitmq producer port
        # 5. rabbitmq producer use-ssl set false!!
        # 6. rabbitmq consumer port
        # 7. rabbitmq consumer use-ssl
        # 8. rabbitmq CA certificate file path
        # 9. rabbitmq exchange name
        # 10. rabbitmq vhost
        rabbitmq_grid.attach(
            Gtk.Label(
                label="Hostname",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            0,
            1,
            1,
        )
        rabbitmq_hostname_entry = engine.register_widget(
            Gtk.Entry(
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "rabbitmq_hostname",
        )
        rabbitmq_grid.attach(rabbitmq_hostname_entry, 1, 0, 2, 1)

        rabbitmq_grid.attach(
            Gtk.Label(
                label="Username",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            1,
            1,
            1,
        )
        rabbitmq_username_entry = engine.register_widget(
            Gtk.Entry(
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "rabbitmq_username",
        )
        rabbitmq_grid.attach(rabbitmq_username_entry, 1, 1, 2, 1)

        rabbitmq_grid.attach(
            Gtk.Label(
                label="Password",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            2,
            1,
            1,
        )
        rabbitmq_password_entry = engine.register_widget(
            Gtk.Entry(
                visibility=False,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "rabbitmq_password",
            exportable=False,
        )
        rabbitmq_grid.attach(rabbitmq_password_entry, 1, 2, 2, 1)

        rabbitmq_grid.attach(
            Gtk.Label(
                label="Producer Port",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            3,
            1,
            1,
        )
        rabbitmq_producer_port_spinbutton = engine.register_widget(
            Gtk.SpinButton(
                adjustment=Gtk.Adjustment(
                    lower=500,
                    upper=50000,
                    value=5672,
                    page_size=0,
                    step_increment=1,
                ),
                value=5672,
                update_policy=Gtk.SpinButtonUpdatePolicy.IF_VALID,
                numeric=True,
                climb_rate=1,
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            "rabbitmq_producer_port",
        )
        rabbitmq_grid.attach(rabbitmq_producer_port_spinbutton, 1, 3, 1, 1)
        # Disable until there's a Ceph release that actually supports this...
        # rabbitmq_grid.attach(
        #    engine.register_widget(Gtk.CheckButton(
        #        label="Use SSL",
        #        halign=Gtk.Align.END, valign=Gtk.Align.CENTER,
        #        hexpand=False, vexpand=False,
        #    ), 'rabbitmq_producer_use_ssl'), 2, 3, 1, 1)

        rabbitmq_grid.attach(
            Gtk.Label(
                label="Consumer Port",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            4,
            1,
            1,
        )
        rabbitmq_consumer_port_spinbutton = engine.register_widget(
            Gtk.SpinButton(
                adjustment=Gtk.Adjustment(
                    lower=500,
                    upper=50000,
                    value=5672,
                    page_size=0,
                    step_increment=1,
                ),
                value=5672,
                update_policy=Gtk.SpinButtonUpdatePolicy.IF_VALID,
                numeric=True,
                climb_rate=1,
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            "rabbitmq_consumer_port",
        )
        rabbitmq_grid.attach(rabbitmq_consumer_port_spinbutton, 1, 4, 1, 1)
        rabbitmq_grid.attach(
            engine.register_widget(
                Gtk.CheckButton(
                    label="Use SSL",
                    halign=Gtk.Align.END,
                    valign=Gtk.Align.CENTER,
                    hexpand=False,
                    vexpand=False,
                ),
                "rabbitmq_consumer_use_ssl",
            ),
            2,
            4,
            1,
            1,
        )

        rabbitmq_grid.attach(
            Gtk.Label(
                label="CA certificate",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            5,
            1,
            1,
        )
        rabbitmq_ca_certificate = engine.register_widget(
            Gtk.FileChooserButton(
                title="Select the CA certificate used to sign the server SSL certificate",
                action=Gtk.FileChooserAction.OPEN,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "rabbitmq_ca_certificate",
        )
        rabbitmq_grid.attach(rabbitmq_ca_certificate, 1, 5, 2, 1)

        rabbitmq_grid.attach(
            Gtk.Label(
                label="Exchange",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            6,
            1,
            1,
        )
        rabbitmq_exchange_entry = engine.register_widget(
            Gtk.Entry(
                text="rfi-file-monitor-ceph",
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "rabbitmq_exchange",
        )
        rabbitmq_grid.attach(rabbitmq_exchange_entry, 1, 6, 2, 1)

        rabbitmq_grid.attach(
            Gtk.Label(
                label="Vhost",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            7,
            1,
            1,
        )
        rabbitmq_vhost_entry = engine.register_widget(
            Gtk.Entry(
                text="/",
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "rabbitmq_vhost",
        )
        rabbitmq_grid.attach(rabbitmq_vhost_entry, 1, 7, 2, 1)

    def _add_horizontal_separator(self):
        self.attach(
            Gtk.Separator(
                orientation=Gtk.Orientation.HORIZONTAL,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=True,
            ),
            0,
            self._row_counter,
            1,
            1,
        )
        self._row_counter += 1



from __future__ import annotations

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib

from ..engine import Engine, EngineThread
from ..file import URL, FileStatus
from ..utils.decorators import exported_filetype, with_pango_docs
from ..utils.exceptions import NotYetRunning

import logging
from urllib.parse import urlparse, unquote_plus
from pathlib import PurePosixPath
from time import time

logger = logging.getLogger(__name__)

ERROR_MSG = (
    "Ensure that the selected file contains valid http and/or https URLs"
)


@with_pango_docs(filename="urls_engine.pango")
@exported_filetype(filetype=URL)
class URLsEngine(Engine):

    NAME = "URLs from File Loader"

    def __init__(self, appwindow):
        super().__init__(appwindow, URLsEngineThread, ERROR_MSG)

        self.attach(Gtk.Label(label="File with URLs"), 0, 0, 1, 1)

        self._file_chooser_button = self.register_widget(
            Gtk.FileChooserButton(
                title="Select a file containing a list of URLs",
                action=Gtk.FileChooserAction.OPEN,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.FILL,
                hexpand=True,
                vexpand=False,
            ),
            "file_with_urls",
        )
        self.attach(self._file_chooser_button, 1, 0, 1, 1)
        self._file_chooser_button.connect(
            "selection-changed", self._file_chooser_button_cb
        )

    def _file_chooser_button_cb(self, button):
        if self.params.file_with_urls is None:
            self._valid = False
        else:
            try:
                with open(self.params.file_with_urls, "r"):
                    pass
                self._valid = True
            except Exception:
                self._valid = False

        self.notify("valid")

    # override parent method, as the thread will have died by the time stop is clicked
    def stop(self):
        if not self._running:
            raise NotYetRunning(
                "The engine needs to be started before it can be stopped."
            )

        self.cleanup()


class URLsEngineThread(EngineThread):
    def run(self):
        # open file
        try:
            with open(self.params.file_with_urls, "r") as f:
                lines = f.readlines()

            # strip lines and ensure they are usable
            lines = list(
                filter(
                    lambda line: not (len(line) == 0 or line.startswith("#")),
                    map(lambda line: line.strip(), lines),
                )
            )

            if len(lines) == 0:
                raise Exception(
                    f"{self.params.file_with_urls} contains no usable URLs"
                )

            # parse URLs
            msg = []
            files = []
            for line in lines:
                parsed = urlparse(line)

                if parsed.scheme not in ("http", "https"):
                    msg.append(
                        f"{line}: only http and https protocols are supported"
                    )
                    continue

                if not parsed.path or not (
                    path := unquote_plus(parsed.path[1:])
                ):
                    msg.append(f"{line}: URL must contain a path")
                    continue

                _file = URL(line, PurePosixPath(path), time(), FileStatus.SAVED)
                files.append(_file)

            if msg:
                msgs = "\n".join(msg)
                raise Exception(
                    f"{self.params.file_with_urls} contains invalid URLs: {msgs}"
                )

            # if we get here, things should be working.
            # close task_window and send files to the queue manager
            GLib.idle_add(
                self._engine.kill_task_window,
                self._task_window,
                priority=GLib.PRIORITY_HIGH,
            )
            GLib.idle_add(
                self._engine._appwindow._queue_manager.add,
                files,
                priority=GLib.PRIORITY_HIGH,
            )
        except Exception as e:
            self._engine.cleanup()
            GLib.idle_add(
                self._engine.abort,
                self._task_window,
                e,
                priority=GLib.PRIORITY_HIGH,
            )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Sequence, Dict, Optional, NamedTuple, Type
import importlib.resources

import yaml
from munch import Munch

from .operation import Operation
from .engine import Engine


class Preference(ABC):
    @abstractmethod
    def __init__(self, key: str, default: Any, description: Optional[str]):
        self._key: str = key
        self._default: Any = default
        self._description = description

    @property
    def key(self) -> str:
        return self._key

    @property
    def default(self) -> Any:
        return self._default

    @property
    def description(self):
        return self._description


class BooleanPreference(Preference):
    def __init__(
        self, key: str, default: bool = False, description: Optional[str] = None
    ):
        super().__init__(key, default, description)


TestBooleanPreference1 = BooleanPreference(
    key="Boolean Pref1",
    default=True,
    description="This is a description for Boolean Pref1",
)

TestBooleanPreference2 = BooleanPreference(key="Boolean Pref2", default=False)

TestBooleanPreference3 = BooleanPreference(
    key="Boolean Pref3", description="This is a description for Boolean Pref3"
)


class ListPreference(Preference):
    def __init__(
        self,
        key: str,
        values: Sequence[str],
        default: Optional[str] = None,
        description: Optional[str] = None,
    ):
        if default and default not in values:
            raise ValueError("default has to be within values array!")
        if not default:
            default = values[0]
        super().__init__(key, default, description)
        self._values = values

    @property
    def values(self) -> Sequence[str]:
        return self._values


TestListPreference1 = ListPreference(
    key="List Pref1",
    values=("Option1", "Option2", "Option3"),
    description="This is a description for List Pref1",
)

TestListPreference2 = ListPreference(
    key="List Pref2",
    values=("Option1", "Option2", "Option3"),
    default="Option3",
)


class DictPreference(Preference):
    def __init__(
        self,
        key: str,
        values: Dict[str, Any],
        default: Optional[str] = None,
        description: Optional[str] = None,
    ):
        if default and default not in values:
            raise ValueError("default has to be within values dict!")
        if not default:
            default = list(values.keys())[0]
        super().__init__(key, default, description)
        self._values = values

    @property
    def values(self) -> Dict[str, Any]:
        return self._values

    @classmethod
    def from_file(
        cls,
        key: str,
        yaml_file,
        default: Optional[str] = None,
        description: Optional[str] = None,
    ):
        with open(yaml_file, "r") as f:
            yaml_dict = yaml.safe_load(stream=f)
        return cls(key, yaml_dict, default, description)


TestDictPreference1 = DictPreference(
    key="Dict Pref1",
    values=dict(
        option1="option1",
        option2=dict(option2="option2"),
        option3=list("option3"),
    ),
    default="option2",
    description="This is a description for Dict Pref1",
)

TestDictPreference2 = DictPreference(
    key="Dict Pref2",
    values=dict(
        option1="option1",
        option2=dict(option2="option2"),
        option3=list("option3"),
    ),
)

with importlib.resources.path(
    "rfi_file_monitor.data", "rfi-instruments.yaml"
) as f:
    TestDictPreference3 = DictPreference.from_file(
        key="Dict Pref3 From File",
        yaml_file=f,
        description="This is a description for Dict Pref3 From File",
    )


class StringPreference(Preference):
    def __init__(
        self, key: str, default: str = "", description: Optional[str] = None
    ):
        super().__init__(key, default, description)


TestStringPreference1 = StringPreference(
    key="String Pref1",
    description="This is a description for String Pref1",
)

TestStringPreference2 = StringPreference(
    key="String Pref2",
    default="String Pref2 default value",
)

AllowedFilePatternsPreference = StringPreference(
    key="Allowed File Patterns",
    description="These comma separated file patterns will be added automatically to the allowed patterns of all operations and engines that support them. Use cautiously!!",
)

IgnoredFilePatternsPreference = StringPreference(
    key="Ignored File Patterns",
    description="These comma separated file patterns will be added automatically to the ignored patterns of all operations and engines that support them.",
    default="*.swp,*.swx,*.DS_Store",
)


class Preferences(NamedTuple):
    settings: Munch[Preference, Any]
    operations: Munch[Type[Operation], bool]
    engines: Munch[Type[Engine], bool]



from __future__ import annotations

import gi

gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")
from gi.repository import Gtk, GLib, Gdk
import yaml

from typing import Final, Dict, Any, List, Optional
import logging
from fnmatch import fnmatch
from time import ctime
from threading import Thread
import inspect
import collections.abc
import traceback

from .utils import (
    PATTERN_PLACEHOLDER_TEXT,
    MONITOR_YAML_VERSION,
    OperationListBox,
)
from .utils.paramswindow import ParamsWindow
from .utils import (
    add_action_entries,
    EXPAND_AND_FILL,
    LongTaskWindow,
    class_in_object_iterable,
)
from .file import FileStatus, File
from .queue_manager import QueueManager
from .engine import Engine
from .operation import Operation
from .preferences import Preferences

logger = logging.getLogger(__name__)


class ApplicationWindow(Gtk.ApplicationWindow):

    ENGINE_ADVANCED_SETTINGS_WINDOW_ATTR = "advanced settings window"

    def __init__(self, force_all=False, **kwargs):
        Gtk.ApplicationWindow.__init__(self, **kwargs)

        self._prefs: Preferences = self.get_property(
            "application"
        ).get_preferences()

        self._preflight_check_metadata: Final[
            Dict[int, Dict[str, Any]]
        ] = dict()
        self._yaml_file: str = None

        self.set_default_size(1000, 1000)

        action_entries = (
            ("save", self.on_save),
            ("save-as", self.on_save_as),
            ("close", self.on_close),
            ("minimize", self.on_minimize),
            ("play", self.on_play),
            ("stop", self.on_stop),
            ("add-operation", self.on_add_operation),
            ("queue-manager", self.on_open_queue_manager),
            ("help-queue-manager", self.on_open_queue_manager_help),
            (
                "status-filter-created",
                self.on_status_filter,
                None,
                GLib.Variant.new_boolean(True),
                FileStatus.CREATED,
            ),
            (
                "status-filter-saved",
                self.on_status_filter,
                None,
                GLib.Variant.new_boolean(True),
                FileStatus.SAVED,
            ),
            (
                "status-filter-queued",
                self.on_status_filter,
                None,
                GLib.Variant.new_boolean(True),
                FileStatus.QUEUED,
            ),
            (
                "status-filter-running",
                self.on_status_filter,
                None,
                GLib.Variant.new_boolean(True),
                FileStatus.RUNNING,
            ),
            (
                "status-filter-success",
                self.on_status_filter,
                None,
                GLib.Variant.new_boolean(True),
                FileStatus.SUCCESS,
            ),
            (
                "status-filter-failure",
                self.on_status_filter,
                None,
                GLib.Variant.new_boolean(True),
                FileStatus.FAILURE,
            ),
        )

        # This doesn't work, which is kind of uncool
        # self.add_action_entries(action_entries)
        for action_entry in action_entries:
            add_action_entries(self, *action_entry)

        self.set_border_width(10)
        main_grid = Gtk.Grid(row_spacing=10, **EXPAND_AND_FILL)
        self.add(main_grid)

        controls_frame = Gtk.Frame(
            label="File Monitor Controls",
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.FILL,
            hexpand=True,
            vexpand=False,
        )
        main_grid.attach(controls_frame, 0, 0, 1, 1)

        controls_grid = Gtk.Grid(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.FILL,
            hexpand=True,
            vexpand=False,
        )

        controls_grid_basic = Gtk.Grid(
            **EXPAND_AND_FILL, border_width=10, column_spacing=5, row_spacing=5
        )
        controls_frame.add(controls_grid)
        controls_grid.attach(controls_grid_basic, 0, 0, 1, 1)
        monitor_play_button = Gtk.Button(
            action_name="win.play",
            image=Gtk.Image(
                icon_name="media-playback-start", icon_size=Gtk.IconSize.DIALOG
            ),
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.END,
            hexpand=False,
            vexpand=False,
        )
        controls_grid_basic.attach(monitor_play_button, 0, 0, 1, 1)
        monitor_stop_button = Gtk.Button(
            action_name="win.stop",
            image=Gtk.Image(
                icon_name="media-playback-stop", icon_size=Gtk.IconSize.DIALOG
            ),
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.START,
            hexpand=False,
            vexpand=False,
        )
        controls_grid_basic.attach(monitor_stop_button, 0, 1, 1, 1)

        # turn the buttons off for now
        # play will become active when operations have been added
        self.lookup_action("play").set_enabled(False)
        self.lookup_action("stop").set_enabled(False)

        # add the notebook with the engines
        self._engines_notebook = Gtk.Notebook(
            **EXPAND_AND_FILL,
            scrollable=True,
        )
        controls_grid_basic.attach(self._engines_notebook, 1, 0, 6, 2)

        self._engines: List[Engine] = list()

        for engine_cls in self.get_property(
            "application"
        ).known_engines.values():
            if not (self._prefs.engines[engine_cls] or force_all):
                continue
            engine = engine_cls(appwindow=self)
            engine_grid = Gtk.Grid(
                **EXPAND_AND_FILL, row_spacing=5, border_width=5
            )
            engine_grid.attach(engine, 0, 0, 1, 1)
            buttons_grid = Gtk.Grid(
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
                column_spacing=5,
            )
            # add button and dialog for advanced settings if necessary
            if (
                engine_cls
                in self.get_property(
                    "application"
                ).engines_advanced_settings_map
            ):
                engine_advanced_settings = self.get_property(
                    "application"
                ).engines_advanced_settings_map[type(engine)](engine)
                title = f"{engine.NAME} Advanced Settings"
                dialog = ParamsWindow(engine_advanced_settings, self, title)
                setattr(
                    engine, self.ENGINE_ADVANCED_SETTINGS_WINDOW_ATTR, dialog
                )

                advanced_settings_button = Gtk.Button(
                    label="Advanced Settings",
                    halign=Gtk.Align.CENTER,
                    valign=Gtk.Align.CENTER,
                    hexpand=True,
                    vexpand=False,
                )
                buttons_grid.attach(
                    advanced_settings_button, len(buttons_grid), 0, 1, 1
                )
                advanced_settings_button.connect(
                    "clicked",
                    self._engine_advanced_settings_button_clicked_cb,
                    engine,
                )
            if engine_cls in self.get_property("application").pango_docs_map:
                help_button = Gtk.Button(
                    label="Help",
                    halign=Gtk.Align.CENTER,
                    valign=Gtk.Align.CENTER,
                    hexpand=True,
                    vexpand=False,
                )
                buttons_grid.attach(help_button, len(buttons_grid), 0, 1, 1)
                help_button.connect(
                    "clicked", self._engine_help_button_clicked_cb, engine
                )

            # fix layout a bit. Buttons should be grouped and centered
            if buttons_grid_len := len(buttons_grid):
                engine_grid.attach(buttons_grid, 0, 1, 1, 1)
                if buttons_grid_len >= 2:
                    # apparently the children of the grid are listed in LIFO order
                    list(buttons_grid)[0].props.halign = Gtk.Align.START
                    list(buttons_grid)[-1].props.halign = Gtk.Align.END
                if buttons_grid_len >= 3:
                    for widget in list(buttons_grid)[1:-1]:
                        widget.props.hexpand = False
            else:
                del buttons_grid
            self._engines_notebook.append_page(
                engine_grid, Gtk.Label(label=engine_cls.NAME)
            )
            self._engines.append(engine)

        # ensure first engine is active
        self._active_engine: Engine = self._engines[0]
        self._active_engine_valid_handler_id = self._active_engine.connect(
            "notify::valid", self._engine_valid_changed_cb
        )
        self._engines_notebook.props.page = 0
        self._engines_notebook.connect("switch-page", self._switch_page_cb)

        # Add support for adding and removing operations
        controls_grid_basic.attach(
            Gtk.Separator(
                orientation=Gtk.Orientation.HORIZONTAL,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=True,
            ),
            0,
            2,
            7,
            1,
        )

        controls_grid_basic.attach(
            Gtk.Label(
                label="<b>Add operation: </b>",
                use_markup=True,
                halign=Gtk.Align.END,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            3,
            2,
            1,
        )

        self._controls_operations_model = Gtk.ListStore(str, object)
        self._controls_operations_combo = Gtk.ComboBox(
            model=self._controls_operations_model,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )

        renderer = Gtk.CellRendererText()
        self._controls_operations_combo.pack_start(renderer, True)
        self._controls_operations_combo.add_attribute(renderer, "text", 0)
        controls_grid_basic.attach(self._controls_operations_combo, 2, 3, 2, 1)

        add_operation_button = Gtk.Button(
            label="Add",
            action_name="win.add-operation",
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        controls_grid_basic.attach(add_operation_button, 4, 3, 2, 1)

        self._repopulate_available_operations()

        if len(self._controls_operations_model) > 0:
            self._controls_operations_combo.set_active(0)
        else:
            self.lookup_action("add-operation").set_enabled(False)
            self._controls_operations_combo.set_sensitive(False)

        paned = Gtk.Paned(
            wide_handle=True,
            orientation=Gtk.Orientation.VERTICAL,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.FILL,
            hexpand=True,
            vexpand=True,
        )
        main_grid.attach(paned, 0, 1, 1, 1)

        operations_frame = Gtk.Frame(
            label="List of Operations",
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.FILL,
            hexpand=True,
            vexpand=True,
        )
        paned.pack1(operations_frame, resize=True, shrink=False)
        operations_scrolled_window = Gtk.ScrolledWindow(
            hscrollbar_policy=Gtk.PolicyType.NEVER,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.FILL,
            hexpand=True,
            vexpand=True,
        )
        operations_frame.add(operations_scrolled_window)
        self._operations_box = OperationListBox(
            selection_mode=Gtk.SelectionMode.NONE,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.FILL,
            hexpand=True,
            vexpand=True,
        )
        operations_scrolled_window.add(self._operations_box)

        output_frame = Gtk.Frame(
            label="Processing Queue",
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.FILL,
            hexpand=True,
            vexpand=True,
        )
        paned.pack2(output_frame, resize=True, shrink=False)

        self._files_tree_model = Gtk.TreeStore(
            str,  # filename, relative to monitored directory
            float,  # epoch time
            int,  # status as code
            str,  # operation name
            float,  # operation progress
            str,  # operation progress as string
            str,  # background color
            str,  # error message
        )

        self._files_tree_model_filter = Gtk.TreeModelFilter(
            child_model=self._files_tree_model
        )
        self._files_tree_model_filter.set_visible_func(
            self._files_tree_model_visible_func
        )

        output_grid = Gtk.Grid(**EXPAND_AND_FILL, row_spacing=2, border_width=2)

        filters_grid = Gtk.Grid(
            border_width=5,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            column_spacing=5,
        )
        output_grid.attach(filters_grid, 0, 0, 1, 1)

        label = Gtk.Label(
            label="Queue Filters:",
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        filters_grid.attach(label, 0, 0, 1, 1)
        state_filters_button = Gtk.Button(
            label="Status",
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        filters_grid.attach(state_filters_button, 1, 0, 1, 1)

        state_filter_popover = Gtk.Popover.new_from_model(
            state_filters_button,
            self.get_property("application").filter_popover_menu,
        )
        state_filters_button.connect(
            "clicked", self._state_filters_button_clicked, state_filter_popover
        )

        label = Gtk.Label(
            label="Name:",
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        filters_grid.attach(label, 2, 0, 1, 1)
        self._name_filter_entry = Gtk.Entry(
            placeholder_text=PATTERN_PLACEHOLDER_TEXT,
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        filters_grid.attach(self._name_filter_entry, 3, 0, 1, 1)
        self._name_filter_entry.connect(
            "changed", self._name_filter_entry_changed
        )

        queue_manager_button = Gtk.Button(
            label="Queue Manager",
            action_name="win.queue-manager",
            halign=Gtk.Align.END,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        filters_grid.attach(queue_manager_button, 4, 0, 1, 1)
        self._queue_manager = QueueManager(self)
        self._queue_manager_window = ParamsWindow(
            self._queue_manager, self, "Queue Manager"
        )

        help_queue_manager_button = Gtk.Button(
            label="Help",
            action_name="win.help-queue-manager",
            halign=Gtk.Align.END,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        filters_grid.attach(help_queue_manager_button, 5, 0, 1, 1)

        files_frame = Gtk.Frame(border_width=5)
        files_scrolled_window = Gtk.ScrolledWindow(**EXPAND_AND_FILL)
        files_frame.add(files_scrolled_window)
        output_grid.attach(files_frame, 0, 1, 1, 1)
        output_frame.add(output_grid)

        files_tree_view = Gtk.TreeView(
            model=self._files_tree_model_filter, border_width=5
        )
        files_scrolled_window.add(files_tree_view)

        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(
            "Filename", renderer, text=0, cell_background=6
        )
        files_tree_view.append_column(column)

        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn("Created", renderer, cell_background=6)
        column.set_cell_data_func(
            renderer, self._time_cell_data_func, func_data=dict(column=1)
        )
        files_tree_view.append_column(column)

        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn("Status", renderer, cell_background=6)
        column.set_cell_data_func(
            renderer, self._status_cell_data_func, func_data=dict(column=2)
        )
        files_tree_view.append_column(column)

        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(
            "Operation", renderer, text=3, cell_background=6
        )
        files_tree_view.append_column(column)

        renderer = Gtk.CellRendererProgress()
        column = Gtk.TreeViewColumn(
            "Progress", renderer, value=4, text=5, cell_background=6
        )
        files_tree_view.append_column(column)

        files_tree_view.set_tooltip_column(column=7)

        # add status bar
        self._status_grid = Gtk.Grid(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            column_homogeneous=True,
            column_spacing=5,
        )
        output_grid.attach(self._status_grid, 0, 2, 1, 1)
        statuses = (
            "Total",
            "Created",
            "Saved",
            "Queued",
            "Running",
            "Success",
            "Failure",
            "Removed from list",
        )
        for _index, _status in enumerate(statuses):
            self._status_grid.attach(
                Gtk.Label(label=f"{_status}: 0"), _index, 0, 1, 1
            )

        # connect delete-event signal handler
        self.connect("delete-event", self._delete_event_cb)

    def _delete_event_dialog_timeout(self, dialog):
        if self._active_engine.props.running:
            return GLib.SOURCE_CONTINUE

        dialog.response(Gtk.ResponseType.CLOSE)

        return GLib.SOURCE_REMOVE

    def _delete_event_killer(self, engine, _):
        if engine.props.running is False:
            self.destroy()

    def _delete_event_cb(self, window, event):
        # If nothing is running, just close it down
        if (
            not self._active_engine.props.running
            and not self._queue_manager.get_property("running")
        ):
            return False

        # else, pop up a dialog asking for confirmation
        dialog = Gtk.MessageDialog(
            buttons=Gtk.ButtonsType.OK_CANCEL,
            message_type=Gtk.MessageType.QUESTION,
            text="The monitor engine is still running!",
            secondary_text="Are you sure you want to close this window?",
            transient_for=self,
            modal=True,
            destroy_with_parent=True,
        )

        source_id = GLib.timeout_add_seconds(
            1, self._delete_event_dialog_timeout, dialog
        )
        rv = dialog.run()
        dialog.destroy()
        GLib.source_remove(source_id)

        if rv in (Gtk.ResponseType.CANCEL, Gtk.ResponseType.DELETE_EVENT):
            return True
        elif rv in (Gtk.ResponseType.OK, Gtk.ResponseType.CLOSE):
            if (
                self._active_engine.props.running
                and self._queue_manager.get_property("running")
            ):
                # hookup signal to active_engine running property
                self._queue_manager.connect(
                    "notify::running", self._delete_event_killer
                )
                # stop it manually
                self._active_engine.stop()
                return True
            return False

    def _switch_page_cb(self, notebook, page, page_num):
        self._active_engine.disconnect(self._active_engine_valid_handler_id)
        self._active_engine = self._engines[page_num]
        self._repopulate_available_operations()
        self._update_monitor_switch_sensitivity()
        self._active_engine_valid_handler_id = self._active_engine.connect(
            "notify::valid", self._engine_valid_changed_cb
        )

    def _repopulate_available_operations(self):
        # get active engine
        filetype_cls = self.get_property(
            "application"
        ).engines_exported_filetype_map[type(self._active_engine)]
        operation_cls_list = []

        for _cls in filetype_cls.__mro__:
            if _cls is File:
                break
            for _op in self.get_property(
                "application"
            ).filetypes_supported_operations_map[_cls]:
                if self._prefs.operations[_op]:
                    operation_cls_list.append(_op)

        operation_cls_list.sort(key=lambda operation: operation.NAME)

        self._controls_operations_model.clear()
        for _class in operation_cls_list:
            self._controls_operations_model.append([_class.NAME, _class])

        if len(self._controls_operations_model):
            self._controls_operations_combo.props.active = 0
            self.lookup_action("add-operation").set_enabled(True)
            self._controls_operations_combo.set_sensitive(True)
        else:
            self.lookup_action("add-operation").set_enabled(False)
            self._controls_operations_combo.set_sensitive(False)

    def _update_monitor_switch_sensitivity(self):
        if len(self._operations_box) == 0:
            self.lookup_action("play").set_enabled(False)
            self._engines_notebook.props.show_tabs = True
        else:
            self._engines_notebook.props.show_tabs = False
            if self._active_engine.props.valid:
                self.lookup_action("play").set_enabled(True)
            else:
                self.lookup_action("play").set_enabled(False)

    def _files_tree_model_visible_func(self, model, iter, data):
        # Children should always be shown when the parent is visible
        if model.iter_parent(iter) is not None:
            return True

        status = FileStatus(model[iter][2])

        if status not in (
            FileStatus.CREATED,
            FileStatus.SAVED,
            FileStatus.QUEUED,
            FileStatus.RUNNING,
            FileStatus.SUCCESS,
            FileStatus.FAILURE,
        ):
            return False

        try:
            status_filter = self.get_action_state(
                f"status-filter-{status.name.lower()}"
            ).get_boolean()
        except AttributeError:
            logger.exception(f"{status=}")
            status_filter = False

        if status_filter is False:
            return False

        pattern = self._name_filter_entry.get_text().strip()

        if not pattern:
            return True
        return fnmatch(model[iter][0], pattern)

    def _engine_advanced_settings_button_clicked_cb(self, button, engine):
        # To avoid problems with the params, we have to reuse the window
        # So when it is closed, it is hidden instead of destroyed.
        dialog = getattr(engine, self.ENGINE_ADVANCED_SETTINGS_WINDOW_ATTR)
        dialog.present()

    def _engine_help_button_clicked_cb(self, button, engine):
        # Reuse window for all engines
        dialog = self.get_property("application").help_window
        dialog.props.transient_for = self
        dialog.select_item(type(engine))
        dialog.present()

    def on_open_queue_manager(self, action, param):
        self._queue_manager_window.present()

    def on_open_queue_manager_help(self, action, param):
        dialog = self.get_property("application").help_window
        dialog.props.transient_for = self
        dialog.select_item(QueueManager)
        dialog.present()

    def on_add_operation(self, action, param):
        _class = self._controls_operations_combo.get_model()[
            self._controls_operations_combo.get_active_iter()
        ][1]
        new_operation = _class(appwindow=self)
        new_operation.index = len(self._operations_box)
        row = Gtk.ListBoxRow(selectable=False, activatable=False)
        row.add(new_operation)
        self._operations_box.insert(row, -1)
        row.show_all()
        self._update_monitor_switch_sensitivity()

    def _remove_operation(self, op_to_remove: Operation):
        for _op in self._operations_box.get_children():
            if _op.get_child() is op_to_remove:
                self._operations_box.remove(_op)
                break
        else:
            logger.error("_remove_operation: op_to_remove not found!")
            return

        self._update_monitor_switch_sensitivity()

        if len(self._operations_box) == 0:
            return

        for op in self._operations_box:
            if op.index > op_to_remove.index:
                op.index = (
                    op.index - 1
                )  # reordering all the indices of the ops.

    def _state_filters_button_clicked(self, button, popover):
        popover.show_all()
        popover.popup()

    def _name_filter_entry_changed(self, entry):
        self._files_tree_model_filter.refilter()

    def _time_cell_data_func(
        self, tree_column, cell, tree_model: Gtk.TreeStore, iter, func_data
    ):
        # we currently dont write a timestamp for the individual operations
        if tree_model.iter_parent(iter) is not None:
            cell.set_property("text", "")
            return
        epoch = tree_model.get_value(iter, func_data["column"])
        date_string = ctime(epoch)
        cell.set_property("text", date_string)

    def _status_cell_data_func(
        self, tree_column, cell, tree_model, iter, func_data
    ):
        status = tree_model.get_value(iter, func_data["column"])
        status_string = str(FileStatus(status))
        cell.set_property("text", status_string)

    def on_minimize(self, action, param):
        self.iconify()

    def on_close(self, action, param):
        self.close()

    def on_status_filter(self, action, param, arg):
        # invert state!
        action.set_state(
            GLib.Variant.new_boolean(not action.get_state().get_boolean())
        )
        self._files_tree_model_filter.refilter()

    def on_play(self, action, param):
        self.lookup_action("play").set_enabled(False)
        task_window = LongTaskWindow(self)
        task_window.set_text("<b>Running preflight check</b>")
        task_window.show()
        watch_cursor = Gdk.Cursor.new_for_display(
            Gdk.Display.get_default(), Gdk.CursorType.WATCH
        )
        task_window.get_window().set_cursor(watch_cursor)

        # Cleanup tree model
        self._files_tree_model.clear()
        PreflightCheckThread(self, task_window).start()

    def on_stop(self, action, param):
        self.lookup_action("stop").set_enabled(False)

        self._active_engine.stop()

    def _preflight_check_cb(
        self, task_window: LongTaskWindow, exception_msgs: Optional[List[str]]
    ):
        task_window.get_window().set_cursor(None)
        task_window.destroy()

        if exception_msgs:
            dialog = Gtk.MessageDialog(
                transient_for=self.get_toplevel(),
                modal=True,
                destroy_with_parent=True,
                message_type=Gtk.MessageType.ERROR,
                buttons=Gtk.ButtonsType.CLOSE,
                text="Operation configuration error(s) found",
                secondary_text="\n".join(exception_msgs),
            )
            dialog.run()
            dialog.destroy()
            self.lookup_action("play").set_enabled(True)
            return

        # Launch the queue manager and the engine
        self._active_engine.set_sensitive(False)
        self._queue_manager.set_sensitive(False)

        for operation in self._operations_box:
            operation.set_sensitive(False)

        self._queue_manager_running_changed_handler_id = (
            self._queue_manager.connect(
                "notify::running", self._queue_manager_running_changed
            )
        )
        self._active_engine_running_changed_handler_id = (
            self._active_engine.connect(
                "notify::running", self._active_engine_running_changed
            )
        )
        self._queue_manager.start()

    def _queue_manager_running_changed(self, queue_manager, param):
        if queue_manager.props.running:
            self._active_engine.start()
        else:
            # at this point everything should have stopped, except for Jobs that still need to finish and cannot be killed
            self._active_engine.set_sensitive(True)
            self._queue_manager.set_sensitive(True)

            for operation in self._operations_box:
                operation.set_sensitive(True)

            self._active_engine.disconnect(
                self._active_engine_running_changed_handler_id
            )
            self._queue_manager.disconnect(
                self._queue_manager_running_changed_handler_id
            )
            del self._active_engine_running_changed_handler_id
            del self._queue_manager_running_changed_handler_id
            self.lookup_action("play").set_enabled(True)

    def _active_engine_running_changed(self, active_engine, param):
        if active_engine.props.running:
            # at this point things should really be running.
            self.lookup_action("stop").set_enabled(True)
            self.get_property(
                "application"
            ).google_analytics_context.send_event(
                "RUN-ENGINE", active_engine.NAME
            )
        else:
            self.lookup_action("stop").set_enabled(False)
            self._queue_manager.stop()

    def _engine_valid_changed_cb(self, engine, param):
        self._update_monitor_switch_sensitivity()

    def _write_to_yaml(self):
        yaml_dict = dict(
            version=MONITOR_YAML_VERSION,
            active_engine=self._active_engine.NAME,
            queue_manager=self._queue_manager.exportable_params,
            operations=[
                dict(name=op.NAME, params=op.exportable_params)
                for op in self._operations_box
            ],
            engines=[
                dict(name=engine.NAME, params=engine.exportable_params)
                for engine in self._engines
            ],
        )
        with open(self._yaml_file, "w") as f:
            yaml.safe_dump(yaml_dict, f)

    def on_save(self, action, param):
        if self._yaml_file is None:
            self.on_save_as(action, param)
        try:
            self._write_to_yaml()
        except Exception as e:
            dialog = Gtk.MessageDialog(
                transient_for=self,
                modal=True,
                destroy_with_parent=True,
                message_type=Gtk.MessageType.ERROR,
                buttons=Gtk.ButtonsType.CLOSE,
                text=f"Could not write to {self._yaml_file}",
                secondary_text=str(e),
            )
            dialog.run()
            dialog.destroy()

    def on_save_as(self, action, param):
        # open a file chooser dialog
        dialog = Gtk.FileChooserNative(
            modal=True,
            title="Save monitor configuration to YAML file",
            transient_for=self,
            action=Gtk.FileChooserAction.SAVE,
        )
        filter = Gtk.FileFilter()
        filter.add_pattern("*.yml")
        filter.add_pattern("*.yaml")
        filter.set_name("YAML file")
        dialog.add_filter(filter)

        if dialog.run() == Gtk.ResponseType.ACCEPT:
            self._yaml_file = dialog.get_filename()
            dialog.destroy()
            # ensure filename ends in .yaml or .yml
            if not self._yaml_file.endswith(
                ".yml"
            ) and not self._yaml_file.endswith(".yaml"):
                self._yaml_file += ".yml"
            try:
                self._write_to_yaml()
            except Exception as e:
                dialog = Gtk.MessageDialog(
                    transient_for=self,
                    modal=True,
                    destroy_with_parent=True,
                    message_type=Gtk.MessageType.ERROR,
                    buttons=Gtk.ButtonsType.CLOSE,
                    text=f"Could not write to {self._yaml_file}",
                    secondary_text=str(e),
                )
                dialog.run()
                dialog.destroy()
        else:
            dialog.destroy()

    def load_from_yaml_dict(self, yaml_dict: dict):

        # active_engine
        active_engine = yaml_dict["active_engine"]
        queue_manager = yaml_dict["queue_manager"]
        operations = yaml_dict["operations"]
        engines = yaml_dict["engines"]

        for engine in engines:
            # find corresponding engine in appwindow
            engine_match = next(
                e for e in self._engines if e.NAME == engine["name"]
            )
            engine_match.update_from_dict(engine["params"])
            if active_engine == engine_match.NAME:
                active_engine_index = self._engines.index(engine_match)
                break
        else:
            logger.error(
                f"Active engine {active_engine} not found in list of engines"
            )
            return

        self._queue_manager.update_from_dict(queue_manager)

        # add the operations
        if isinstance(operations, collections.abc.Sequence):
            for op in operations:
                for _class in self.get_property(
                    "application"
                ).known_operations.values():
                    if op["name"] == _class.NAME:
                        new_operation = _class(appwindow=self)
                        new_operation.index = len(self._operations_box)
                        self._operations_box.insert(new_operation, -1)
                        new_operation.update_from_dict(op["params"])
                        new_operation.show_all()
                        break
                else:
                    logger.error(
                        f"load_from_yaml_dict: no match found for operation {op['name']}"
                    )

        self._engines_notebook.set_current_page(active_engine_index)
        self._update_monitor_switch_sensitivity()

    @property
    def preflight_check_metadata(self) -> dict:
        return self._preflight_check_metadata

    @property
    def queue_manager(self) -> QueueManager:
        return self._queue_manager

    @property
    def active_engine(self) -> Engine:
        return self._active_engine


class PreflightCheckThread(Thread):
    def __init__(
        self, appwindow: ApplicationWindow, task_window: LongTaskWindow
    ):
        super().__init__()
        self._appwindow = appwindow
        self._task_window = task_window

    def run(self):
        exception_msgs = []
        self._appwindow._preflight_check_metadata.clear()

        for index, operation in enumerate(self._appwindow._operations_box):
            if index > 0 and hasattr(operation, "PREREQUISITES"):
                preceding_ops = list(self._appwindow._operations_box)[0:index]
                preceding_ops_str = map(lambda op: op.NAME, preceding_ops)
                for prerequisite in getattr(operation, "PREREQUISITES"):
                    if (
                        inspect.isclass(prerequisite)
                        and issubclass(prerequisite, Operation)
                        and not class_in_object_iterable(
                            preceding_ops, prerequisite
                        )
                    ):

                        exception_msgs.append(
                            f"* {prerequisite.NAME} must precede {operation.NAME}"
                        )
                        break
                    elif (
                        isinstance(prerequisite, str)
                        and prerequisite not in preceding_ops_str
                    ):

                        exception_msgs.append(
                            f"* {prerequisite} must precede {operation.NAME}"
                        )
                        break
            elif hasattr(operation, "PREREQUISITES"):
                prereq = ", ".join(
                    map(
                        lambda x: x if isinstance(x, str) else x.NAME,
                        getattr(operation, "PREREQUISITES"),
                    )
                )
                exception_msgs.append(
                    f"* Operation {operation.NAME} must be proceded by {prereq}"
                )

            try:
                operation.preflight_check()
            except Exception as e:
                logger.info(
                    f"Exception caught from {operation.NAME}: {traceback.format_exc()}"
                )
                exception_msgs.append(f"* {operation.NAME}: " + str(e))

        if exception_msgs:
            for operation in self._appwindow._operations_box:
                operation.preflight_cleanup(success=False)

            GLib.idle_add(
                self._appwindow._preflight_check_cb,
                self._task_window,
                exception_msgs,
                priority=GLib.PRIORITY_DEFAULT_IDLE,
            )
            return

        for operation in self._appwindow._operations_box:
            operation.preflight_cleanup(success=True)

        GLib.idle_add(
            self._appwindow._preflight_check_cb,
            self._task_window,
            None,
            priority=GLib.PRIORITY_DEFAULT_IDLE,
        )



import requests

# pylint: disable=import-error
from requests.packages.urllib3.util.retry import Retry

from ..file import URL
from ..queue_manager import QueueManager
from .s3_downloader import S3DownloaderOperation
from ..utils import get_file_creation_timestamp, TimeoutHTTPAdapter
from ..utils.decorators import supported_filetypes, with_pango_docs
from ..utils.exceptions import SkippedOperation

import logging
import datetime
from dateutil.parser import parse as parsedate
from pathlib import Path

logger = logging.getLogger(__name__)

SESSION = requests.Session()
adapter = TimeoutHTTPAdapter(
    timeout=2.5,
    max_retries=Retry(
        backoff_factor=1,
        status_forcelist=[429, 500, 502, 503, 504],
        method_whitelist=[
            "HEAD",
            "GET",
            "PUT",
            "DELETE",
            "OPTIONS",
            "TRACE",
            "POST",
        ],
        total=5,
    ),
    pool_maxsize=QueueManager.MAX_JOBS,
)
SESSION.mount("https://", adapter)
SESSION.mount("http://", adapter)


@supported_filetypes(filetypes=URL)
@with_pango_docs(filename="url_downloader.pango")
class UrlDownloaderOperation(S3DownloaderOperation):

    NAME = "URL Downloader"
    CHUNK_SIZE = 1024 * 50

    def run(self, file: URL):

        # Check if file already exists in the destination folder
        destination = Path(
            self.params.download_destination, *file.relative_filename.parts
        )

        # try a get request to get the response header only
        response = SESSION.get(file.filename, allow_redirects=True, stream=True)
        try:
            response.raise_for_status()
        except Exception as e:
            return str(e)

        try:
            url_size = int(response.headers["Content-Length"])
        except Exception:
            url_size = None

        if destination.exists():
            # get the filesize
            try:
                file_size = destination.stat().st_size
                if url_size == file_size:
                    url_time = parsedate(
                        response.headers["Last-Modified"]
                    ).astimezone()
                    file_time = datetime.datetime.fromtimestamp(
                        get_file_creation_timestamp(destination)
                    ).astimezone()
                    if url_time < file_time:
                        raise SkippedOperation(
                            "File has been downloaded already"
                        )
            except SkippedOperation:
                raise
            except Exception:
                pass
        else:
            destination.parent.mkdir(parents=True, exist_ok=True)

        with destination.open("wb") as f:
            downloaded_bytes = 0
            last_percentage = 0

            for chunk in response.iter_content(chunk_size=self.CHUNK_SIZE):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)

                    if url_size:
                        downloaded_bytes += self.CHUNK_SIZE
                        percentage = (downloaded_bytes / url_size) * 100
                        if int(percentage) > last_percentage:
                            last_percentage = int(percentage)
                            file.update_progressbar(self.index, last_percentage)



from __future__ import annotations
from rfi_file_monitor.utils import ExitableThread

from gi.repository import Gio

from ..engine_advanced_settings import EngineAdvancedSettings
from ..engine import Engine
from ..utils.exceptions import SkippedOperation
from ..file import File, RegularFile, Directory, WeightedRegularFile, FileStatus
from ..operation import Operation

from typing import Type, Union, Sequence, Callable, Optional
import logging
import inspect
from pathlib import Path
import collections.abc
import functools
import threading

logger = logging.getLogger(__name__)

_app = Gio.Application.get_default()


def with_pango_docs(filename: str):
    """Decorator for engines and operations, used to set the name of the file
    whose contents should be used to populate the associated Help dialog.
    Provide the basename of the file only, and make sure it is placed in a folder
    called `docs`, which must be a subfolder within the folder containing the engine or operation"""

    def _with_pango_docs(cls: Type[Union[Operation, Engine]]):
        if _app is None:
            return cls
        if not issubclass(cls, Operation) and not issubclass(cls, Engine):
            logger.error(
                f"with_pango_cos can only be used to decorate classes that extend Engine, Operation or QueueManager"
            )
            return cls
        try:
            module = inspect.getmodule(cls)
            if not module:
                raise Exception(f"module for class {cls.NAME} not found")
            contents = (
                Path(module.__file__)
                .parent.joinpath("docs", filename)
                .read_text()
            )
        except Exception:
            logger.exception(
                f"with_pango_docs: could not open {filename} for reading"
            )
        else:
            _app.pango_docs_map[cls] = contents
        return cls

    return _with_pango_docs


def with_advanced_settings(
    engine_advanced_settings: Type[EngineAdvancedSettings],
):
    """Decorator for Engine classes, to be used when some of their
    settings have been delegated to an advanced settings window
    OPTIONAL."""

    def _with_advanced_settings(cls: Type[Engine]):
        if _app is None:
            return cls
        if not issubclass(cls, Engine):
            logger.error(
                f"with_advanced_settings can only be used to decorate classes that extend Engine"
            )
            return cls
        _app.engines_advanced_settings_map[cls] = engine_advanced_settings
        return cls

    return _with_advanced_settings


# This may need to be changed later, if an engine can record multiple filetypes...
def exported_filetype(filetype: Type[File]):
    """Decorator for Engine classes that declares which filetype
    it will be looking out for. MANDATORY. Without this decorator,
    the engine cannot be tied to operations."""

    def _exported_filetype(cls: Type[Engine]):
        if _app is None:
            return cls
        if not issubclass(cls, Engine):
            logger.error(
                f"exported_filetype can only be used to decorate classes that extend Engine"
            )
            return cls
        _app.engines_exported_filetype_map[cls] = filetype
        return cls

    return _exported_filetype


def supported_filetypes(filetypes: Union[Type[File], Sequence[Type[File]]]):
    """Decorator for Operation classes that should be used to declare
    which filetype(s) it supports. OPTIONAL. If unused, then the operation
    will be assumed to support regular files only!"""

    def _supported_filetypes(cls: Type[Operation]):
        if _app is None:
            return cls
        if not issubclass(cls, Operation):
            logger.error(
                f"supported_filetypes can only be used to decorate classes that extend Operation"
            )
            return cls
        if isinstance(filetypes, collections.abc.Sequence):
            _filetypes = filetypes
        else:
            _filetypes = [filetypes]
        for filetype in _filetypes:
            if filetype in _app.filetypes_supported_operations_map:
                _app.filetypes_supported_operations_map[filetype].append(cls)
            else:
                _app.filetypes_supported_operations_map[filetype] = [cls]
        return cls

    return _supported_filetypes


# We are using filesize to determine weight in progressbar changes
# However, if the combined filesize is 0, then the file index will be used instead.
def add_directory_support(run: Callable[[Operation, File], Optional[str]]):
    @functools.wraps(run)
    def wrapper(self: Operation, file: File):
        current_thread = threading.current_thread()

        if isinstance(file, RegularFile):
            return run(self, file)
        elif isinstance(file, Directory):
            # get all files contained with Directory, as well as their sizes
            _path = Path(file.filename)
            _parent = _path.parent
            total_size = file.total_size
            size_seen = 0

            file.operation_metadata[self.index] = {}
            for file_index, (filename, size) in enumerate(file):
                # abort if job has been cancelled
                if (
                    isinstance(current_thread, ExitableThread)
                    and current_thread.should_exit
                ):
                    return str("Thread killed")

                if total_size == 0:
                    offset = file_index / len(file)
                    weight = 1 / len(file)
                else:
                    offset = size_seen / total_size
                    size_seen += size
                    weight = size / total_size

                _file = WeightedRegularFile(
                    filename,
                    Path(filename).relative_to(_parent),
                    0,
                    FileStatus.CREATED,
                    offset,
                    weight,
                )
                # reuse the row_reference to ensure the progress bars are updated
                _file.row_reference = file.row_reference

                # run the wrapped method, and do the usual exception and return value handling
                try:
                    rv = run(self, _file)
                    if rv:
                        return rv
                except SkippedOperation:
                    pass

                if self.index in _file.operation_metadata:
                    file.operation_metadata[self.index][
                        _file.filename
                    ] = _file.operation_metadata[self.index]
            return None
        else:
            raise NotImplementedError(f"{type(file)} is currently unsupported")

    return wrapper
